/*********************************************************************
    This file is part of Siworin. Copyright © 2021 by Ruud Harmsen.

    Siworin is a simple word indexer and local search engine. See
    https://rudhar.com/sfreview/siworin/#siw04 and
    https://rudhar.com/sfreview/siworin/toolsrc/siworin .

    Siworin is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    Siworin is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Siworin.  If not, see <https://www.gnu.org/licenses/>.
  ********************************************************************/

/* Siworin - simple word indexer
   8, 11 September 2021,
   by Ruud Harmsen

   5 and 6 November 2022:
   Fix a bug. For example when three words matched, adding
   an extra search word, occurring in the same file but too
   far away, made the whole match disappear because of the
   too great distance. But of course even though the 4-word
   match does not hold, the 3-word match still does!

   5 September 2024:
   Measures for special case when two or more of the search
   words are indentical. Dutch example that actually occurs
   on my website rudhar.com: dat dat dat.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "siworin.h"
#include "siworin-match.h"
#include "siworin-ranking.h"
#include "siworin-displ.h"
#include "siworin-files.h"
#include "siworin-config.h"

static int numelts_hfile_wordvector = 0;
static struct hfile_wordvector_s *wordbase = NULL;

static char *fgets_expand (FILE *fp);
static void setup_matchvectors (char *buf);
static void find_matches (void);
static int prxsrch_compar (const void *a, const void *b);
static int match_found (int minim_match, int maxim_dista);
static void SpecialCaseEqualSrchwrds (void);
static void free_all (void);

#ifdef DEBUG
static void report_vectors (int simple);
#endif

int siworin_findlocsfromfiles (FILE *fp, int combined)
{
   char *buf;

   rewind(fp);
   /* Read one line from the file into a malloc'ed buffer,
      large enough to hold the whole line. */
   while (buf = fgets_expand(fp), !feof(fp))
   {
      setup_matchvectors(buf);
      if (!combined)
      {
         find_matches();
         free_all();
      }
   }
   free(buf);
#ifdef DEBUG2
   report_vectors(0);
#endif
   if (combined)
   {
      find_matches();
      free_all();
   }
   return 0;
}

/* Read a line of unknown length, in several parts, allocating
   more space as we go. */
#define CHUNK 127
static char *fgets_expand (FILE *fp)
{
   char *ret, *buf = NULL;
   /* Bufsize does NOT include room for the closing nullbyte!
      So allocate one byte more each time. */
   int bufsiz = 0;
   /* Index where the result of the last fgets can be tested.
      If it contains .. that means ..:
      '\n' The portion read fitted in the buffer exactly
           (followed by newline and nullbyte)
      '\0' Nullbyte is still there after previous initialisation:
           the portion read was shorter than the buffer.
      other: The portion read (followed by nullbyte, but not ending
           in newline) doesn't complete the full line: they is more
           left for reading. */
   int totest;
   /* Index where the next fgets (if any) should begin storing its
      data. */
   int tofill = 0;

   do {
      if (!(buf = realloc(buf, bufsiz + CHUNK + 1)))
         SIWORIN_FAIL("realloc");
      bufsiz += CHUNK;
      totest = bufsiz - 1;
      buf[totest] = '\0';
      ret = fgets(buf + tofill, CHUNK + 1, fp);
      tofill += CHUNK;
   } while (ret && buf[totest] && buf[totest] != '\n');

   return buf;
}

static void setup_matchvectors (char *buf)
{
   char *cp;
   int i = numelts_hfile_wordvector;
   int consumed = 0;

#ifdef DEBUG2
   fprintf(stdout, "<p>%s line %4d: Handling this line:\n%s\n",
      __FILE__, __LINE__, buf);
#endif

   wordbase = realloc(wordbase, (i + 1) * sizeof *wordbase);
   if (!wordbase)
      SIWORIN_FAIL("realloc");
   numelts_hfile_wordvector++;

   if ((cp = strchr(buf, SIWORIN_SEP)) == NULL)
      SIWORIN_FAIL("strchr");

   if (!(wordbase[i].srch_word = malloc(cp - buf + 1)))
      SIWORIN_FAIL("malloc");
   sprintf(wordbase[i].srch_word, "%.*s", (int)(cp - buf), buf);

   for (wordbase[i].vec = NULL,
        wordbase[i].currind = 0,
        wordbase[i].numelts_vec = 0;               /* for-loop init */
        1;                                         /* for-loop condition */
        cp += consumed, wordbase[i].numelts_vec++) /* for-loop increment */
   {
      wordbase[i].vec = realloc(wordbase[i].vec,
         (wordbase[i].numelts_vec + 1) * sizeof (struct hfilepos_s));
      if (!wordbase[i].vec)
         SIWORIN_FAIL("realloc");
      if (sscanf(cp, SIWORIN_FORMAT2"%n",
         &wordbase[i].vec[wordbase[i].numelts_vec].pathline,
         &wordbase[i].vec[wordbase[i].numelts_vec].offseth,
         &consumed) != 2)
         break;
   }
}

static void find_matches (void)
{
   int minim_match = atoi(siworin_config_value(SIWORIN_MINIM_MATCH));
   /* In bytes. */
   int maxim_dista = atoi(siworin_config_value(SIWORIN_MAXIM_DISTA));

   if (!wordbase)
      /* Can happen if none of the searchwords were found in
         the word index. */
      return;

   if (minim_match > numelts_hfile_wordvector)
       minim_match = numelts_hfile_wordvector;

   siworin_ranking_init(numelts_hfile_wordvector);

   while (1)
   {
      /* An optimisation would be possible, avoiding to sort the whole
         array if the first two elements are already in increasing order.
         I tested this, and found the effect on performance is negligible,
         even in the presence of a very frequent search word which in fact
         doesn't contribute to any matches, so it has to be incremented
         out of the way sequentially, re-sorting the array unnecessarily
         each time.
         So I removed the optimisation, and kept things 'dumbsmart' and
         simple. */
      qsort(wordbase,
         numelts_hfile_wordvector, sizeof *wordbase, prxsrch_compar);
      #ifdef DEBUG2
      fprintf(stdout, "<p>%s line %4d:\n", __FILE__, __LINE__);
      report_vectors(1);
      #endif

      SpecialCaseEqualSrchwrds();

      if (wordbase[numelts_hfile_wordvector-1].currind >=
          wordbase[numelts_hfile_wordvector-1].numelts_vec)
         break;

      if (match_found(minim_match, maxim_dista))
         if (siworin_ranking_add1hit(wordbase, numelts_hfile_wordvector))
            SIWORIN_FAIL("ranking");

      wordbase[0].currind++;
   }
   siworin_ranking_finalise(numelts_hfile_wordvector);
}

static int prxsrch_compar (const void *av, const void *bv)
{
   const struct hfile_wordvector_s *a = av;
   const struct hfile_wordvector_s *b = bv;

   /* Special case: both elements to be compared are exhausted.
      Return a consistent value, to avoid vacillation. */
   if (a->currind >= a->numelts_vec &&
       b->currind >= b->numelts_vec)
      return -1;

   if (a->currind >= a->numelts_vec)
      return +1;
   if (b->currind >= b->numelts_vec)
      return -1;

   if (a->vec[a->currind].pathline != b->vec[b->currind].pathline)
      return a->vec[a->currind].pathline >
             b->vec[b->currind].pathline ? 1 : -1;

   return a->vec[a->currind].offseth >
          b->vec[b->currind].offseth ? 1 : -1;
}

static int match_found (int minim_match, int maxim_dista)
{
   int i, samefile = 0;
   unsigned long prev = wordbase[0].vec[wordbase[0].currind].pathline;

   /* Check for at least mimim_match equal file index numbers */
   for (i = 1;
        i < numelts_hfile_wordvector &&
        wordbase[i].currind < wordbase[i].numelts_vec;
        i++)
   {
      if (wordbase[i].vec[wordbase[i].currind].pathline != prev)
         break;
   }
   samefile = i;
   if (samefile < minim_match)
      return 0;

   #ifdef DEBUG2
   fprintf(stdout, "<p>%s line %4d, %d words match, minimum was %d\n",
      __FILE__, __LINE__, samefile, minim_match);
   #endif

   /* Are the distances OK? */
   prev = wordbase[0].vec[wordbase[0].currind].offseth;
   for (i = 1; i < samefile; i++)
   {
      /* No match if too great a byte distance between words. */
      if ((long)wordbase[i].vec[wordbase[i].currind].offseth - (long)prev
        > (long)maxim_dista)
      {
         #ifdef DEBUG2
         fprintf(stdout,
            "<p>Comparing %d-%d: offsets %lx(%ld) - %lx(%ld) = %ld (max. %d)\n",
            i, i - 1,
            wordbase[i].vec[wordbase[i].currind].offseth,
            wordbase[i].vec[wordbase[i].currind].offseth,
            prev, prev,
            wordbase[i].vec[wordbase[i].currind].offseth - prev,
            maxim_dista);
         fprintf(stdout, "No match\n");
         #endif

         /* Check for a special case: if without this too distant
            word, the previous group of words did make a match, it
            should be reported. */
         if (i >= minim_match)
         {
            #ifdef DEBUG2
               fprintf(stdout, "This is a match anyway!\n");
            #endif
            return 1;
         }
         return 0;
      }
      prev = wordbase[i].vec[wordbase[i].currind].offseth;
   }
#ifdef DEBUG2
   fprintf(stdout, "This is a match!\n");
#endif
   return 1;
}

/* Special case when two or more search words are the same,
   the standard algorithm doesn't properly handle that. Need
   to force the vector indices on different occurrences of
   the searchwords then. */
static void SpecialCaseEqualSrchwrds (void)
{
   int EqualPresent;

   do {
      EqualPresent = 0;

      for (int i = 0; i < numelts_hfile_wordvector - 1; i++)
      {
         if (wordbase[i + 1].currind < wordbase[i + 1].numelts_vec &&
             wordbase[i    ].currind < wordbase[i    ].numelts_vec &&
             wordbase[i + 1].vec[wordbase[i + 1].currind].pathline ==
             wordbase[i    ].vec[wordbase[i    ].currind].pathline &&
             wordbase[i + 1].vec[wordbase[i + 1].currind].offseth ==
             wordbase[i    ].vec[wordbase[i    ].currind].offseth)
         {
            wordbase[i + 1].currind++;
            EqualPresent = 1;
            qsort(wordbase,
               numelts_hfile_wordvector, sizeof *wordbase, prxsrch_compar);
            break;
         }
      }
   } while (EqualPresent);
}

static void free_all (void)
{
   for (int i = 0; i < numelts_hfile_wordvector; i++)
   {
      free(wordbase[i].srch_word);
      free(wordbase[i].vec);
   }
   free(wordbase);
   wordbase = NULL;
   numelts_hfile_wordvector = 0;
}

#ifdef DEBUG
static void report_vectors (int simple)
{
   int i, j;

   fprintf(stdout, "<p>%s line %4d:\n", __FILE__, __LINE__);
   fprintf(stdout, "<p>%s line %4d, %d</p>\n", __FILE__, __LINE__,
      (int)numelts_hfile_wordvector);

   for (i = 0; i < numelts_hfile_wordvector; i++)
   {
      fprintf(stdout, "<br>%s\telts %4d ", wordbase[i].srch_word,
                                           wordbase[i].numelts_vec);
      if (simple)
      {
         fprintf(stdout, "<p>%s line %4d:\n", __FILE__, __LINE__);
         if (wordbase[i].currind >= wordbase[i].numelts_vec)
            fprintf(stdout, "!---!---\n");
         else
            fprintf(stdout, "!%lx!%lx %ld\t%p\n",
               wordbase[i].vec[wordbase[i].currind].pathline,
               wordbase[i].vec[wordbase[i].currind].offseth,
               wordbase[i].vec[wordbase[i].currind].offseth,
               (void *)wordbase[i].vec);
      }
      else
      {
         fprintf(stdout, "<p>%s line %4d:\n", __FILE__, __LINE__);
         for (j = 0; j < wordbase[i].numelts_vec; j++)
         {
            fprintf(stdout, "!%lx!%lx",
               wordbase[i].vec[j].pathline, wordbase[i].vec[j].offseth);
         }
         fprintf(stdout, "<p>%s line %4d:\n", __FILE__, __LINE__);
         fprintf(stdout, "\n");
      }
      fprintf(stdout, "<p>%s line %4d:\n", __FILE__, __LINE__);
   }
   fprintf(stdout, "\n");

   fprintf(stdout, "<p>%s line %4d:\n", __FILE__, __LINE__);
}
#endif

#ifdef MAINMATCH
int main (int argc, char **argv)
{
   FILE *fp = fopen(argc < 2 ? "fnd" : argv[1], "r");
   if (!fp)
      SIWORIN_FAIL("fopen");

   siworin_args_to_config_path(argv);
   if (siworin_fileopen_paths(
       siworin_config_value(SIWORIN_CONFIG_INDEX_FILES_PREFIX),
       "r", "r"))
      SIWORIN_FAIL("fopens");

   if (siworin_fileopen_words(
       siworin_config_value(SIWORIN_CONFIG_INDEX_FILES_PREFIX),
       "r", "r"))
      SIWORIN_FAIL("fopens");

   siworin_findlocsfromfiles(fp, 0);

   fclose(fp);
   siworin_fileopen_close();

   return 0;
}
#endif
